/* $Id: pm2free.c,v 1.5 1999/12/18 15:12:28 becka Exp $
******************************************************************************

** 
 * KGI - PM2 fbcon derived acceleration driver
 * Copyright (c) 1998-1999 Ilario Nardinocchi (nardinoc@CS.UniBO.IT)
 * Driver used for port under KGI license with permission of Ilario
 * Nardinocchi.
 * Copyright (c) 1999 Andreas Beck - port to KGI
**
     
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Andreas_Beck
#define	DRIVER_NAME	"Permedia 2 fbcon accelerator driver"
#define	DRIVER_REV	"$Revision: 1.5 $"

/*
**	KGI includes
*/

#include <kgi/module.h>
#include <kgi/types.h>
#include <kgi/kgi_commands.h>

#include <3dlabs/pm2free.h>

/*#define NUMPINGPONGPAGES 2*/

/* We don't need the suggest stuff for kgicon */
#undef NEED_SUGGEST_FEAT

#ifdef NEED_SUGGEST_FEAT
const char *sugg_default[KGIGT_LAST][5] = {
	{ "generic-text-16","generic-ramdac",NULL,NULL,NULL },
	{ "generic-text-32","generic-ramdac",NULL,NULL,NULL },
	{ "generic-linear-1","generic-ramdac",NULL,NULL,NULL },
	{ "vendor-IBM-vga-4bpp",NULL,NULL,NULL,NULL},
	{ "generic-linear-8","generic-ioctl","generic-ramdac",NULL,NULL},
	{ "generic-linear-15",NULL,NULL,NULL,NULL},
	{ "generic-linear-16","generic-ioctl",NULL,NULL},
	{ "generic-linear-24","generic-ioctl","generic-gamma",NULL,NULL},
	{ "generic-linear-32","generic-ioctl","generic-gamma",NULL,NULL}
};
#endif /* NEED_SUGGEST_FEAT */

/*
**	KGI module interface
*/
int kgim_accel_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev " DRIVER_REV " loaded");
	/* !!! Check for Chipset and Version here !!! */

	return EOK;
}

void kgim_accel_done(struct kgi_display *dpy)
{
	NOTICE1(DRIVER_NAME " removed");
}
/*
 * Check for a valid graphic mode.
 */
int kgim_accel_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
	enum kgi_tm_cmd *cmd)
{
	switch (*cmd) {

	case CMD_CHECK:
		/* 
		 * Check for a valid mode - we grok all.
		*/
		return EOK;

	default:
		return EOK;
	}
}

/*
 * Setup the accelerator engine.
 */
void kgim_accel_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
/*	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset); */

	return;	/* chipset driver does it all */
}

/*
***************************************************************************
 Actual acceleration functions

  Most of them are inlined to gain speed.
  As they are only called from one place this doesn't waste any space.
***************************************************************************
*/

static void wait_pm2(struct chipset_private* i) {

	WAIT_FIFO(i, 1);
	pm2_WR(i, PM2R_SYNC, 0);
	DEFRW();
	do {
		while (pm2_RD(i, PM2R_OUT_FIFO_WORDS)==0);
		DEFR();
	} while (pm2_RD(i, PM2R_OUT_FIFO)!=PM2TAG(PM2R_SYNC));
}

/*
 * block operation. copy=0: rectangle fill, copy=1: rectangle copy.
 */
static void pm2fb_block_op(struct chipset_private* i, int copy,
					kgis32 xsrc, kgis32 ysrc,
					kgis32 x, kgis32 y, kgis32 w, kgis32 h ) {

	if (!w || !h)
		return;
	WAIT_FIFO(i, 6);
	pm2_WR(i, PM2R_CONFIG,	PM2F_CONFIG_FB_WRITE_ENABLE|
				PM2F_CONFIG_FB_READ_SOURCE_ENABLE);
	pm2_WR(i, PM2R_FB_PIXEL_OFFSET, 0);
	if (copy)
		pm2_WR(i, PM2R_FB_SOURCE_DELTA,	((ysrc-y)&0xfff)<<16|
							((xsrc-x)&0xfff));
	pm2_WR(i, PM2R_RECTANGLE_ORIGIN, (y<<16)|x);
	pm2_WR(i, PM2R_RECTANGLE_SIZE, (h<<16)|w);
	DEFW();
	pm2_WR(i, PM2R_RENDER,	PM2F_RENDER_RECTANGLE|
				(x<xsrc?PM2F_INCREASE_X:0)|
				(y<ysrc?PM2F_INCREASE_Y:0)|
				(copy?0:PM2F_RENDER_FASTFILL));
	wait_pm2(i);
}

/* Set the accelerator's clipping rectangle
 */
static inline void pm2free_setclip(struct kgi_display *dpy,
				   struct kgi_gc *gc)
{
/* FIXME !
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);

	mem_out32((gc->cliptl.x & 0x07FF) | (((gc->clipbr.x-1) & 0x07FF) << 16),
		  CXBNDRY);
	mem_out32((dpy->mode.request.virt.x*gc->cliptl.y) & 0x00FFFFFF,
		  YTOP);
	mem_out32((dpy->mode.request.virt.x*(gc->clipbr.y-1)) & 0x00FFFFFF,
		  YBOT);
*/
}

static kgiu32 mapcolor(struct kgi_display *dpy,kgiu32 color)
{
       	switch(dpy->mode.request.graphtype) {
		case KGIGT_8BIT:
			color&=0xff;
			color|=(color<<8);
			color|=(color<<16);
			return color;
		case KGIGT_16BIT:
			color&=0xffff;
			color|=(color<<16);
			return color;
		case KGIGT_32BIT:
		default:
			return color;
       	}
	return color;
}

/* Update GC if neccesary
 */
static inline void pm2free_updategc(struct kgi_graphic *gr)
{
	struct kgi_display *dpy = gr->dev.dpy;
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
	kgi_gc *gc = gr->gc;

	if (gc->version != gc->oldversion) {
		pm2free_setclip(gr->dev.dpy, gc);
        	
		WAIT_FIFO(private, 2);
		pm2_WR(private, PM2R_FB_BLOCK_COLOR          , mapcolor(dpy,gc->fg_color) );
		pm2_WR(private, PM2R_COLOR_DDA_CONSTANT_COLOR, mapcolor(dpy,gc->fg_color) );

		gc->oldversion = gc->version;
	}
}

/* CopyBox using the accelerator
 */
static inline int pm2free_copybox(struct kgi_graphic *gr, 
				  struct kgi_cbox *boxp)
{	
	struct kgi_display *dpy = gr->dev.dpy;
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
	kgiu32 x = boxp->x, y = boxp->y, w = boxp->w, h = boxp->h,
		dstx = boxp->x2, dsty = boxp->y2;

       	if(dpy->mode.request.graphtype==KGIGT_24BIT)
		return -E(ACCEL,NOSUP_ALWAYS_MMAP);
	pm2fb_block_op(private, 1, x, y, dstx, dsty, w, h );
	return EOK;
}

static void pm2fb_line_op(struct chipset_private* i, 
					kgis32 x1, kgis32 y1,
					kgis32 x2, kgis32 y2 ) {

	if (x1==x2) {
		if (y1<y2)
			pm2fb_block_op(i, 0, x1, y1, x1, y1, y2-y1+1, 1 );
		else
			pm2fb_block_op(i, 0, x1, y2, x1, y2, y1-y2+1, 1 );
		return;
	}
	WAIT_FIFO(i, 3);
	pm2_WR(i, PM2R_CONFIG,	PM2F_CONFIG_FB_WRITE_ENABLE);
	pm2_WR(i, PM2R_FB_PIXEL_OFFSET, 0);
	pm2_WR(i, PM2R_COLOR_DDA_MODE, PM2R_COLOR_DDA_MODE_ENABLE | PM2R_COLOR_DDA_MODE_FLAT );
	WAIT_FIFO(i, 6);
	pm2_WR(i, PM2R_START_X_DOM, (x1<<16));
	pm2_WR(i, PM2R_D_X_DOM,      (1<<16));
	pm2_WR(i, PM2R_START_Y    , (y1<<16));
	pm2_WR(i, PM2R_D_Y        , ((y2-y1)<<16)/(x2-x1) );
	pm2_WR(i, PM2R_COUNT      , (x2>x1) ? (x2-x1) : (x1-x2) );
	DEFW();
	pm2_WR(i, PM2R_RENDER,	PM2F_RENDER_LINE);
	wait_pm2(i);
}

static void pm2fb_triangle_op(struct chipset_private* i, 
					kgis32 x1, kgis32 y1,
					kgis32 x2, kgis32 y2,
					kgis32 x3, kgis32 y3 ) {

	kgis32 hlp;

	if (y1>y2) {	hlp=x1;x1=x2;x2=hlp;	hlp=y1;y1=y2;y2=hlp;	}
	if (y1>y3) {	hlp=x1;x1=x3;x3=hlp;	hlp=y1;y1=y3;y3=hlp;	}
	if (y2>y3) {	hlp=x2;x2=x3;x3=hlp;	hlp=y2;y2=y3;y3=hlp;	}
	
	if (y1==y3) return;	/* Fixme - degenerate case. */

	WAIT_FIFO(i, 3);
	pm2_WR(i, PM2R_CONFIG,	PM2F_CONFIG_FB_WRITE_ENABLE);
	pm2_WR(i, PM2R_FB_PIXEL_OFFSET, 0);
	pm2_WR(i, PM2R_COLOR_DDA_MODE, PM2R_COLOR_DDA_MODE_ENABLE | PM2R_COLOR_DDA_MODE_FLAT );

	if (y1<y2) {	/* top half is not degenerated. Draw it */

		WAIT_FIFO(i, 8);
		pm2_WR(i, PM2R_START_X_DOM, (x1<<16));
		pm2_WR(i, PM2R_D_X_DOM    , ((x3-x1)<<16)/(y3-y1) );
		pm2_WR(i, PM2R_START_X_SUB, (x1<<16));
		pm2_WR(i, PM2R_D_X_SUB    , ((x2-x1)<<16)/(y2-y1) );
		pm2_WR(i, PM2R_START_Y    , (y1<<16));
		pm2_WR(i, PM2R_D_Y        , (1<<16) );
		pm2_WR(i, PM2R_COUNT      , (y2-y1)+1 );
		DEFW();
		pm2_WR(i, PM2R_RENDER,	PM2F_RENDER_TRAPEZOID );
#if 0
		if (y2<y3) { 	/* bottom half is not degenerated. Draw it */
			WAIT_FIFO(i, 3);
			pm2_WR(i, PM2R_START_X_SUB, (x2<<16));
			pm2_WR(i, PM2R_D_X_SUB,     ((x3-x2)<<16)/(y3-y2) );
			DEFW();
			pm2_WR(i, PM2R_CONTINUE_NEW_SUB, (y3-y2) );
		}
	} else {	/* degenerate case - top is flat - we just draw from the bottom up. */
#else
	}
	if (y2<y3) {	/* bottom half is not degenerated. Draw it */
#endif
		WAIT_FIFO(i, 8);
		pm2_WR(i, PM2R_START_X_DOM, (x3<<16));
		pm2_WR(i, PM2R_D_X_DOM    , ((x3-x1)<<16)/(y1-y3) );
		pm2_WR(i, PM2R_START_X_SUB, (x3<<16));
		pm2_WR(i, PM2R_D_X_SUB    , ((x3-x2)<<16)/(y2-y3) );
		pm2_WR(i, PM2R_START_Y    , (y3<<16));
		pm2_WR(i, PM2R_D_Y        , (-1<<16) );
		pm2_WR(i, PM2R_COUNT      , (y3-y2)+((y1<y2) ? 0 : 1) );
		DEFW();
		pm2_WR(i, PM2R_RENDER,	PM2F_RENDER_TRAPEZOID );
	}
	wait_pm2(i);
}

/* Draw a box with the accelerator
 */
static inline int pm2free_drawbox(struct kgi_graphic *gr, struct kgi_cbox *boxp)
{
	struct kgi_display *dpy = gr->dev.dpy;
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
	kgiu32 x = boxp->x, y = boxp->y, w = boxp->w, h = boxp->h;

       	if(dpy->mode.request.graphtype==KGIGT_24BIT)
		return -E(ACCEL,NOSUP_ALWAYS_MMAP);
	pm2fb_block_op(private, 0, x, y, x, y, w, h );
	return EOK;
}

/* Draw a line using the accelerator
 */
static inline int pm2free_drawline(struct kgi_graphic *gr, struct kgi_line *line)
{
	struct kgi_display *dpy = gr->dev.dpy;
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
	kgiu32 x1 = line->x, y1 = line->y, x2 = line->x2, y2 = line->y2;

       	if(dpy->mode.request.graphtype==KGIGT_24BIT)
		return -E(ACCEL,NOSUP_ALWAYS_MMAP);
	pm2fb_line_op(private, x1, y1, x2, y2);
	return EOK;
}

/* Draw a triangle using the accelerator
 */
static inline int pm2free_drawtriangle(struct kgi_graphic *gr, struct kgi_triangle *triangle)
{
	struct kgi_display *dpy = gr->dev.dpy;
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
	kgiu32  x1 = triangle->x1, y1 = triangle->y1, 
		x2 = triangle->x2, y2 = triangle->y2,
		x3 = triangle->x3, y3 = triangle->y3;

       	if(dpy->mode.request.graphtype==KGIGT_24BIT)
		return -E(ACCEL,NOSUP_ALWAYS_MMAP);
	pm2fb_triangle_op(private, x1, y1, x2, y2, x3, y3);
	return EOK;
}

int kgim_accel_command (struct kgi_graphic *gr, kgi_uint cmd, kgi_uint arg)
{
	if (!(gr->dev.state & KGI_MAPPED_ON_DISPLAY) ) {
		/* Don't accel inactive screen! */
	        return -E(ACCEL, NOSUP_MODE_MMAP);
	}

	pm2free_updategc(gr);
       	
	switch (cmd) { 
	case ACCEL_COPYBOX:
		return pm2free_copybox(gr, (struct kgi_cbox *) gr->io_buf);
	   
	case ACCEL_DRAWLINE:
		return pm2free_drawline(gr, (struct kgi_line *) gr->io_buf);
	   
	case ACCEL_DRAWBOX:
		return pm2free_drawbox(gr, gr->io_buf);

	case ACCEL_DRAWTRIANGLE:
		return pm2free_drawtriangle(gr, (struct kgi_triangle *)gr->io_buf);

#ifdef NEED_SUGGEST_FEAT
	case ACCEL_GETSUGGEST:
	{
		kgi_suggest *sugg;
		const char *name;

		sugg = (kgi_suggest *)(gr->io_buf);
		gr->io_res = gr->io_buf;
      
		if (sugg->id < 0 || sugg->id > 4) {

			return -E(ACCEL, BOGUS);
		}
      
		name = sugg_default[gr->dev.mode.graphtype][sugg->id];
		if (!name) {

			return -E(ACCEL, BOGUS);
		}
/*		if (gr->dev.mode.graphtype == KGIGT_8BIT &&
			gr->dev.mode.virt.x == 320 &&
			gr->dev.mode.virt.y == 200 &&
			strcmp("generic-ioctl", name) == 0) {
		   
			return -E(ACCEL, BOGUS);
		}*/
      
		strcpy(sugg->name, name);
		sugg->args[0] = 0;
		sugg->id++;	/* Next is RAMDAC */
		if (sugg->id >= 4 ||
		    sugg_default[gr->dev.mode.graphtype][sugg->id] == NULL) {

			sugg->id = 0;
		}
		return 0;
	}
#endif /* NEED_SUGGEST_FEAT */

	default: 
	  
	        return -E(ACCEL,NOSUP_ALWAYS_MMAP); 
	}
}



